@@ -1,4 +1,5 @@
-# $Id: fileasserts.rb 41859 2013-07-09 09:03:32Z knu $
+# frozen_string_literal: true
+# $Id: fileasserts.rb 57275 2017-01-07 02:14:07Z kazu $
 
 module Test
   module Unit
@@ -86,6 +87,24 @@ EOT
 File modes expected to be equal:
  <#{'%0*o' % [width, mode1]}>: "#{file1}"
  <#{'%0*o' % [width, mode2]}>: "#{file2}"
+EOT
+      end
+
+      def assert_ownership_group(expected, file)
+        actual = File.stat(file).gid
+        assert expected == actual, <<EOT
+File group ownership of "#{file}" unexpected:
+ Expected: <#{expected}>
+   Actual: <#{actual}>
+EOT
+      end
+
+      def assert_ownership_user(expected, file)
+        actual = File.stat(file).uid
+        assert expected == actual, <<EOT
+File user ownership of "#{file}" unexpected:
+ Expected: <#{expected}>
+   Actual: <#{actual}>
 EOT
       end
     end
